/*
 * NetcraftToolbar - Main Extension Code
 *
 * Copyright (C) 2014 Netcraft Ltd
 * http://toolbar.netcraft.com/ 
 *
 */


if (!com) var com={};
if (!com.netcraft) com.netcraft={};
if (!com.netcraft.toolbar) com.netcraft.toolbar={};

com.netcraft.toolbar.phishtank = {
    constants: com.netcraft.toolbar.constants,
    debug: com.netcraft.toolbar.debug,
    previous_url: null,
    DNScanceller: null,
    
    _rootObj: null,
    _root: function () {
        if (this._rootObj == null) {
            this._rootObj        = top;
            this._rootObj.lookup = new Object;
            this._rootObj.xss    = this.buildPatterns(this.URLarray(this.constants.PT_XSSDATURL));
            this._rootObj.lb     = this.buildPatterns(this.URLarray(this.constants.PT_LBDATURL));
        }
        
        return this._rootObj;
    },
    
    QueryInterface: function(aIID) {
        if (aIID.equals(Components.interfaces.nsIWebProgressListener) ||
            aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
            aIID.equals(Components.interfaces.nsISupports))
        return this;
    throw Components.results.NS_NOINTERFACE;
    },
    
    onLocationChange: function(aProgress, aRequest, aURI) {
        return 0;
    },

    onStateChange: function(aProgress, aRequest, aStateFlags, aStatus) {
        var nsIWebProgressListener = Components.interfaces.nsIWebProgressListener;

        if (aStateFlags & nsIWebProgressListener.STATE_IS_DOCUMENT &&
            !(aStateFlags & nsIWebProgressListener.STATE_STOP)) {
            if(aRequest && aRequest instanceof
                Components.interfaces.nsIChannel) {

                if (aRequest.URI.spec == this.previous_url) {
                    return 0;
                }

                var toolbar = document.getElementById(this.constants.PT_TOOLBARID);
                if ((toolbar != null) && toolbar.collapsed) {
                    this.displayHoster("");
                    return 0;
                }
                this.changeToolbar(aRequest.URI, aProgress.DOMWindow, aRequest);
            }
        }
        return 0;
    },
    onProgressChange: function() { return 0; },
    onStatusChange: function() { return 0; },
    onSecurityChange: function() { return 0; },
    onLinkIconAvailable: function() { return 0; },

    changeToolbar: function (aURI, aDOMWindow, aRequest) {
        
        /* If we don't have a suitable scheme, we should display nothing
         * rather than not changing the text in the toolbar
         */
        this.previous_url = aURI.spec;
        if (this.DNScanceller) {
            this.DNScanceller();
            this.DNScanceller = null;
        }

        switch(aURI.scheme) {
            case "http":
            case "https":
            case "ftp":
                break;
            default:
                this.displayHoster("", aDOMWindow);
                return;
        }

        var key = aURI.scheme + "://" + aURI.asciiHost;
        var url = aURI.asciiSpec;

        if (this._root().xss &&
            this.checkPatterns(this._root().xss, this.canonicaliseURL(url)) &&
            this.yesno("Suspected Cross-Site Scripting (XSS) detected!", this.constants.PT_XSSMSG, url, aRequest)
        ) {
            this.displayHoster("XSS URL blocked!", aDOMWindow);
            this.replaceLocation(this.constants.PT_XSSBLOCKEDURL, aDOMWindow);
            return;
        }

        if (this._root().lb &&
            this.checkPatterns(this._root().lb, url) &&
            this.yesno("Suspicious URL detected!", this.constants.PT_LBMSG, url, aRequest)
        ) {
            this.displayHoster("Suspicious URL blocked!", aDOMWindow);
            this.replaceLocation(this.constants.PT_LBBLOCKEDURL, aDOMWindow);
            return;
        }

        if (this.isCached(key)) {
            this.updateIframe(this._root().lookup[key][0], url, false, aDOMWindow);
        } else if (com.netcraft.toolbar.options.isOffline()) {
            this.displayHoster("offline");
            return;
        } else {
            var check_url = this.constants.PT_CHECKURL + aURI.scheme + "://" + aURI.asciiHost;
            if (aDOMWindow.top == aDOMWindow) {
                this.DNScanceller = com.netcraft.toolbar.dns.dnsResolve(aURI.host, onReturnedIP);
            } else {
                check_url += "/embedded";
            }
            this.fetchInfo(key, check_url, url, true, true, aDOMWindow, aRequest);
        }

        function onReturnedIP(ip) {
            var phishtank = com.netcraft.toolbar.phishtank;

            if (ip) {
                if (com.netcraft.toolbar.dns.unroutableIP(ip)) {
                    phishtank.displayHoster("[" + daddy.constants.PT_NOINFO + "]", aDOMWindow);
                    return;
                } else {
                    check_url += "/" + com.netcraft.toolbar.dns.convertIP(ip);
                }
                if (aDOMWindow.top != aDOMWindow) {
                    check_url += "/embedded";
                }
                check_url += "/info";
                phishtank.fetchInfo(key, check_url, url, true, false, aDOMWindow, aRequest);
            } else {
                phishtank.displayHoster("[" + daddy.constants.PT_NOINFO + "]", aDOMWindow);
                return;
            }
        }
        return;
    },

    fetchInfo: function (key,chkurl,url, first_time, just_patterns, aDOMWindow, aRequest) {
        var daddy = this;
        var req = new XMLHttpRequest();
        req.open("GET", chkurl, true);
        req.key = key;
        req.onreadystatechange = function() {
            if (req.readyState == 4) {
                var redirect;
                if (redirect = req.getResponseHeader("Location") && first_time) {
                    daddy.fetchInfo(key, redirect, url, false, just_patterns, aDOMWindow, aRequest);
                } else {
                    daddy.updateIframe(req, url, just_patterns, aDOMWindow, aRequest);
                }
            } else {
                if (!just_patterns) {
                    daddy.displayHoster("Checking " + key + "....!".substr(0,req.readyState+1), aDOMWindow);
                }
            }
        }
        req.send(null);
        return true;
    },

    updateIframe: function (req, url, just_patterns, aDOMWindow, aRequest) {
        
        var resp = this.parseResponse(req, url, just_patterns);
        if (!just_patterns) {
            this.displayHoster(resp.descr, aDOMWindow);
        }
        if (resp.patterns) {
            var patArray = this.headerArray(atob(resp.patterns));
            if (this.checkPatterns(this.buildPatterns(patArray), this.removeURLHash(decodeURI(url))) &&
                this.yesno("Phishing Site detected!", this.constants.PT_PHISHINGMSG, url, aRequest))
            {
                this.displayHoster("Phishing URL blocked!", aDOMWindow);
                this.replaceLocation(this.constants.PT_PHISHINGBLOCKEDURL, aDOMWindow);
            }
        }
    },

    parseResponse: function (req, url, just_patterns) {
        var resp;
        var descr;
        var patterns;
        if (req.getResponseHeader("Content-Type").match("application/json")) {
            switch (req.status) {
                case 200:
                    resp = JSON.parse(req.responseText);
                    if (!just_patterns) {
                        this.addCache(req);
                        descr = this.constructDisplay(resp, url);
                    }
                    if (resp.patterns) {
                        patterns = resp.patterns;
                    }
                    break;
    
                case 0:
                    descr = "[" + this.constants.PT_SERVERERROR + ": " + this.constants.PT_NETWORKERROR + "]";
                    break;
                case 204:
                    descr = "[" + this.constants.PT_NOINFO + "]";
                    break;
                case 422:
                    resp = JSON.parse(req.responseText);
                    descr = "[" + resp.error + "]";
                    break;
            
                default:
                    descr = "[" + this.constants.PT_SERVERERROR + ": " + req.statusText + "]";
                    break;
            }
        } else if (req.getResponseHeader("Content-Type").match("text/html")) {
            resp = req.responseText;
            if (req.status === 200 && resp.match("<title>Site Access Error</title>")) {
                descr = "[" + this.constants.PT_RATELIMITED + "]";
            } else {
                descr = "[" + this.constants.PT_SERVERERROR + ": " + req.statusText + "]";
            }
        } else {
            descr = "[" + this.constants.PT_SERVERERROR + ": " + req.statusText + "]";
        }

        return { descr:descr, patterns:patterns };
    },

    constructDisplay: function (resp, url) {
        var flag = resp.country;
        if (flag && flag.match(/^\w\w$/)) {
            flag = '<img src="chrome://netcrafttoolbar/content/flags/' +
                flag.toLowerCase() +
                '.gif"/>' +
                ' [' + flag + ']';
        } else {
            flag = '';
        }

        var risk = resp.risk;
        if (risk && risk >= 0 && risk <= 10) {
            var barLength = Math.round(risk * this.constants.PT_RISKBARLEN / 10);
            risk =
                '<span style="position: relative;" title="">' +
                '<a href="http://toolbar.netcraft.com/help/faq/index.html#riskrating" ' + 
                    'title="Risk Rating: ' + risk + '"' +
                '>' +
                  '<img style="' + this.constants.PT_ABSPOS + '" ' +
                'border="0" ' +
                'src="chrome://netcrafttoolbar/content/risk.png" ' +
                    'title="Risk Rating: ' + risk + '"' +
                  '/>' +
                  '<img style="' + this.constants.PT_ABSPOS +
                'clip: rect(0px ' + barLength + 'px auto 0px);" ' +
                'border="0" ' +
                'src="chrome://netcrafttoolbar/content/bar.png" ' +
                  '/>' + 
                '</a>' +
                '</span>';
        } else {
            risk = '';
        }

        var report_url = this.constants.PT_SITEREPORTURL + '?url=' +
                url.match(/^\w+:\/\/[^\/]+/);

        var report = '<a href="' + report_url + '">';

        var since = resp.firstseen;
        if (since && since.match(/^\w+ \d{4}$/)) {
            since = 'Since: ' + 
                report +
                since + '</a>';
        } else {
            since = 'Since: ' +
                report +
                'New Site</a>';
        }

        var rank = resp.rank;
        var topsites = resp.topsites;
        if (rank && rank > 0 && topsites && topsites.match(/^[0-9A-F]{28}$/)) {
            rank = 'Rank: <a href="' +
                this.constants.PT_TOPSITESURL + '?s=' +
                topsites + '#' + rank + '">' +
                rank + '</a>';
        } else {
            rank = 'Rank: -';
        }

        var pfs = '';
        if (resp.pfs == 0 || resp.pfs == 1) {
            var pfs_color, pfs_icon;
            if (resp.pfs == 1) {
                pfs_color = '#007700';
                pfs_icon = '&#x2714;';
            } else if (resp.pfs == 0) {
                pfs_color = '#ff3333';
                pfs_icon = '&#x2718;';
            }

            pfs = '<a href="' + this.constants.PT_PFSURL + '">PFS</a>: '
                + '<a href="' + this.constants.PT_PFSURL + '"'
                + ' style="color: ' + pfs_color + ' !important; text-decoration: none !important;">'
                + pfs_icon + '</a>';
        }

        var sslv3 = '';
        if (resp.uses_sslv3 == 0 || resp.uses_sslv3 == 1) {
            var sslv3_icon = '<img id="sslv3-icon" src="chrome://netcrafttoolbar/content/warning.png" />';

            if (resp.uses_sslv3 == 1) {
                sslv3 = '<a href="' + this.constants.PT_POODLEURL + '">SSLv3</a>: '
                      + sslv3_icon;
            }
            else if (resp.uses_sslv3 == 0) {
            }
        }

        var heartbleed = '';
        if (resp.heartbleed == 'unsafe') {
            var heartbleed_message = '';
            if (resp.heartbleed_message) {
                heartbleed_message = resp.heartbleed_message;
                heartbleed_message = heartbleed_message.replace(/</g,'&lt;');
                heartbleed_message = heartbleed_message.replace(/>/g,'&gt;');
                heartbleed_message = heartbleed_message.replace(/"/g,'&quot;');
                heartbleed_message = heartbleed_message.replace(/&(?!\w+;)/g,'&amp;');
            }
            heartbleed = '<span title="' + heartbleed_message + '">'
                       + '<a href="' + this.constants.PT_HEARTBLEEDURL + '">Heartbleed</a>: '
                       + '<a href="' + this.constants.PT_HEARTBLEEDURL + '">'
                       + '<img id="heartbleed" src="chrome://netcrafttoolbar/content/heartbleed.png" />'
                       + '</a></span>';
        }

        var netblock = resp.netblock;
        var hoster = resp.hoster;
        if (netblock && hoster) {
            netblock = escape(netblock);
            hoster = hoster.replace(/</g,'&lt;');
            hoster = hoster.replace(/>/g,'&gt;');
            hoster = hoster.replace(/"/g,'&quot;');
            hoster = hoster.replace(/&(?!\w+;)/g,'&amp;');

            netblock = '<a href="' +
                this.constants.PT_NETBLOCKURL + '?q=' +
                netblock + '">' +
                hoster + '</a>';
        } else {
            netblock = '';
        }

        descr = risk +
                '<nobr style="margin-left: 60px;">' +
                since + ' ' +
                rank + ' ' +
                pfs + ' ' +
                sslv3 + ' ' +
                heartbleed + ' ' +
                report + 'Site Report</a> ' +
                flag + ' ' +
                netblock +
                '</nobr>';

        descr = descr.replace(/\<a /g, '<a onclick="com.netcraft.toolbar.phishtank.goTo(this.href); return(false);" ');
        return descr;
    },

    displayHoster: function (htmlText, aDOMWindow) {

        /* make sure we are in a top level (i.e. not an iframe or frame)
         * before we modify the contents of the toolbar
         */
        if (aDOMWindow.top != aDOMWindow) {
            return 0;
        }

        /* Do not display anything if the event isn't for the currently
         * selected Browser
         */
        if (aDOMWindow &&
            gBrowser.selectedBrowser.contentWindow.location.href != aDOMWindow.location.href) {
            return 0;
        }

        var doc = document.getElementById(this.constants.PT_DESCRIPTIONID);
        while (doc.hasChildNodes())
            doc.removeChild(doc.firstChild);
        this.insertXHTML(doc, htmlText);
    },

    insertXHTML: function(targetElement, xhtml) {
        var parser = new DOMParser();
        var doc = parser.parseFromString('<div xmlns="http://www.w3.org/1999/xhtml">' + xhtml + '</div>', 'application/xhtml+xml');
        
        if (doc.getElementsByTagName("parsererror").length > 0) {
            var error = "XML is not well-formed, displaying unavailable for:\n" + xhtml;
            this.debug.consoleerror(error);
            xhtml = "[" + this.constants.PT_SERVERERROR + ": " + this.constants.PT_UNAVAILABLEERROR + "]";
            doc = parser.parseFromString('<div xmlns="http://www.w3.org/1999/xhtml">' + xhtml + '</div>', 'application/xhtml+xml');
        }
        
        var root = doc.documentElement;
        for (var i = 0; i < root.childNodes.length; i++) {
            targetElement.appendChild(document.importNode(root.childNodes[i], true));
        }
    },

    substURL: function (str,url,encodingNeeded) {
        if(encodingNeeded){
            str = str.replace(/%URL%/, this.formatURL(url.toString()));
        }
        else {
            str = str.replace(/%URL%/, (url.toString()));
        }
        return str.replace(/%UUID%/, com.netcraft.toolbar.options.uuid);
    },
    
    yesno: function (title,str,url,aRequest) {
        /*
         * suspend the request whilst we wait for an answer 
         */ 
         var suspended = false;
        if (aRequest instanceof Components.interfaces.nsIRequest) {
            if(aRequest.isPending()) {
                aRequest.suspend();
                suspended = true;
            }
        }

        var factory = Components.classes["@mozilla.org/prompter;1"].getService(Components.interfaces.nsIPromptFactory);
        var prompt = factory.getPrompt(window, Components.interfaces.nsIPrompt);

        var flags = prompt.BUTTON_TITLE_IS_STRING * prompt.BUTTON_POS_0 +
            prompt.BUTTON_TITLE_IS_STRING * prompt.BUTTON_POS_1 +
            prompt.BUTTON_POS_1_DEFAULT;

        var warning = str;
        str = this.constants.PT_ALERTMSG.replace(/%ACTION%/,str);
        if (url && url.length) {
            str = this.substURL(str, this.shortenURL(url), false);
        }
	    if (warning == this.constants.PT_XSSMSG  || warning == this.constants.PT_LBMSG) {
            str = str.replace(/%PT_REPORTMSG%/, "");
        }
        else {
            str = str.replace(/%PT_REPORTMSG%/, this.constants.PT_REPORTMSG);        
        }

        var returnVal = prompt.confirmEx(title, str, flags, "Yes", "No", null, null, {});
        /* If we still do want to go, let's resume the request */
        if(!returnVal && suspended) {
            aRequest.resume();
        }
        return returnVal;
    },

    openURL: function (aURL) {

        var theTop = top.document.getElementById("content");
        var newtab = theTop.addTab(aURL);

        window.getBrowser().selectedTab = newtab;
    },

    goTo: function (url) {
        var loc = window.content.location;
        url = this.substURL(url,loc,true);
        this.openURL(url);
    },

    /* Network Read */
    URLarray: function (url) {
        var serv = Components.classes["@mozilla.org/network/io-service;1"].
            getService(Components.interfaces.nsIIOService);
        var chan = serv.newChannel(url, null, null);
        var sis = Components.classes["@mozilla.org/scriptableinputstream;1"].
            createInstance(Components.interfaces.nsIScriptableInputStream);
        sis.init(chan.open());

        var buf = "";
        while(sis.available() > 0)
              buf += sis.read(this.constants.PT_BUFSIZ);
        buf = buf.replace(/\s+$/,'');

        sis.close();
        return (buf.length > 0 ? buf.split(/[\r\n]+/) : null);
    },

    headerArray: function (str) {
        str = str.replace(/\s+$/,'');
        return (str.split(/\t/));
    },

    canonicaliseURL: function (url) {
        url = unescape(url.replace(/\+/g, " "));
        url = url.replace(/[\x00-\x08\x0e-\x1f\x7f-\xff]/g, "");
        url = url.replace(/\\/g, "");
        return url;
    },

    removeURLHash: function (url) {
	var ioService = Components.classes["@mozilla.org/network/io-service;1"]
                       .getService(Components.interfaces.nsIIOService);
        var urlParser = ioService.newURI(url, null, null);
        urlParser.ref = null;
        return urlParser.spec;
    },

    shortenURL: function (url) {
        var len = url.length;
        if(len > this.constants.PT_MAXURLLEN + 3) {
            var part_len = Math.round((this.constants.PT_MAXURLLEN - 3) / 4);
            url = url.substring(0, part_len * 3) + "..."
                + url.substring(len - part_len, len);
        }
        return url;        
    },

    formatURL: function (url) {
        url = url.
            replace(/\%/g, "%25").
            replace(/\#/g, "%23").
            replace(/\+/g, "%2b").
            replace(/\&/g, "%26").
            replace(/\?/g, "%3f").
            replace(/\s+/g,'+');

        return url;
    },

    buildPatterns: function (strarray) {
        if (!strarray || strarray.length == 0)
            return null;

        var patterns = new Array;
        for (var s = 0; s < strarray.length; s++) {
            strarray[s] = strarray[s].replace(/\(\?[\w-]+\)/g,"");
            patterns[s] = new RegExp(strarray[s],"i");
        }
        return(patterns.length > 0 ? patterns : null);
    },

    checkPatterns: function (patterns,str) {
        if (!patterns || patterns.length == 0)
            return false;

        for (var pat = 0; pat < patterns.length; pat++) {
            if (patterns[pat].exec(str)) {
            return true;
            }
        }
        return false;
    },

    isCached: function (key) {
        if (!this._root().lookup[key]) {
            return false;
        }
        var d = new Date();
        if (d.getTime() > this._root().lookup[key][1]) {
            d.setTime(this._root().lookup[key][1]);
            delete this._root().lookup[key];
            return false;
        }
        return true;
    },

    addCache: function (req) {
        if (this.isCached(req.key)) {
            return;
        }
            var d = new Date();
        var expires = req.getResponseHeader("Expires");
        if (expires) {
            d.setTime(Date.parse(expires));
        } else {
            d.setTime(d.getTime() + (this.constants.PT_CACHECHECKURL * 1000));
        }
        this._root().lookup[req.key] = [req, d.getTime()];
    },

    /* 
     * Make sure we replace the context of the correct tab
     */
    replaceLocation: function (location, aDOMWindow) {
        aDOMWindow.location.replace(location);
        return;
    },

    register: function(aBrowser) {
        aBrowser.addProgressListener(
            this,
            Components.interfaces.nsIWebProgress.NOTIFY_STATE_DOCUMENT
        );
    },

    deregister: function(aBrowser) {
        aBrowser.removeProgressListener(this);
    }
};

/*
 * Handle registering the listeners per-window per-tab. In essence, we
 * will have one toolbar object per window, and then we have invidual
 * per-tab events being recorded by each per-window toolbar.
 */

com.netcraft.toolbar.listener = {};
com.netcraft.toolbar.listener.tab = {
    load: function(event) {
        var browser = gBrowser.getBrowserForTab(event.target);
        com.netcraft.toolbar.phishtank.register(browser);
    },
    unload: function(event) {
        var browser = gBrowser.getBrowserForTab(event.target);
        com.netcraft.toolbar.phishtank.deregister(browser);
	},
    selectTab: function(event) {
        var browser = gBrowser.selectedBrowser;
        com.netcraft.toolbar.phishtank.changeToolbar(browser.currentURI,browser.contentWindow);
    },
};

com.netcraft.toolbar.listener.window =  {
    load: function(event) {
        var container = gBrowser.tabContainer;
        container.addEventListener("TabOpen", com.netcraft.toolbar.listener.tab.load, false);
        container.addEventListener("TabClose", com.netcraft.toolbar.listener.tab.unload, false);
        container.addEventListener("TabSelect", com.netcraft.toolbar.listener.tab.selectTab, false);

        /*
         * We need to explicitly register the open tabs as they will
         * not be registered by virtue of being opened
         */
        for ( var i = 0; i < gBrowser.browsers.length; i++) {
            var browser = gBrowser.getBrowserAtIndex(i);
            com.netcraft.toolbar.phishtank.register(browser);
        }
    },
    unload: function(event) {
        var container = gBrowser.tabContainer;
        container.removeEventListener("TabOpen", com.netcraft.toolbar.listener.tab.load, false);
        container.removeEventListener("TabClose", com.netcraft.toolbar.listener.tab.unload, false);
        container.removeEventListener("TabSelect", com.netcraft.toolbar.listener.tab.selectTab, false);

    },
};

/*
 * These two listeners are where everything kicks off, we register
 * load event-listeners to trigger on window load events which in turn
 * register handlers for per-tab events.
 */
window.addEventListener('load', com.netcraft.toolbar.listener.window.load, false);
window.addEventListener('unload', com.netcraft.toolbar.listener.window.unload, false);
